%{
#include <stdio.h>
#include <unistd.h>	
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include "nsu.h"
#include "nsu_app.h"
#include "y.tab.h"
	
#define MAX_STRING 1024
static char strbuf[MAX_STRING];
static size_t stridx;

static void
str_add(char const *str, size_t len)
{
        if (len + 1 > MAX_STRING - stridx)
		abend("input string too long");
        memcpy(strbuf + stridx, str, len);
        stridx += len;
}

static void
str_start(char const *str, size_t len)
{
        stridx = 0;
        str_add(str, len);
}

static void
str_esc(int c)
{
	if (stridx == MAX_STRING-1)
		abend("input string too long");
	strbuf[stridx++] = c;
}

static char *
str_stop(void)
{
	return alloc_string(strbuf, stridx);
}
        
static char *
str_new(char const *str, size_t len)
{
        str_start(str, len);
        return str_stop();
}

struct strlist {
	struct strlist *prev;
	char str[1];
};

static struct strlist *strtail;

char *
alloc_string(char const *src, size_t len)
{
	struct strlist *sp = malloc(sizeof(*sp) +  len);
	if (!sp)
		abend_nomem();
	memcpy(sp->str, src, len);
	sp->str[len] = 0;
	sp->prev = strtail;
	strtail = sp;
	return sp->str;
}

void
dealloc_strings(void)
{
	while (strtail) {
	        struct strlist *prev = strtail->prev;
		free(strtail);
		strtail = prev;
        }
}

static int inargc;
static char **inargv;
static int next;
static char *curp;
static int state;

static inline int
is_interactive(void)
{
	return !inargv && isatty(fileno(yyin)) && isatty(1);
}

static int need_prompt = 1;

static inline void
prompt(void)
{
	if (need_prompt) {
		printf("nsu> ");
		fflush(stdout);
	}
}

#define ST_DFL 0
#define ST_QUOTE 0x1
#define ST_DELIM 0x2
#define ST_STOP 0x4

static int
fillbuf(char *buf, size_t max_size)
{
	if (inargv) {
		size_t total = 0;
		char const specials[] = " \"\t\n;";
		char const escapable[] = "\\\"";
  
		while (total < max_size) {
			size_t len;

			if (curp == NULL || *curp == 0) {
				if (state & ST_QUOTE) {
					buf[total++] = '"';
					state &= ~ST_QUOTE;
					continue;
				}
				if (state & ST_STOP)
					break;
				if (next == inargc) {
					buf[total++] = ';';
					state |= ST_STOP;
					break;
				}
				if (next)
					state |= ST_DELIM;
				curp = inargv[next++];
			}
	  
			if (state & ST_DELIM) {
				buf[total++] = ' ';
				state &= ~ST_DELIM;
				continue;
			}

			len = strlen(curp);
			if (!(state & ST_QUOTE)) {
				size_t n = strcspn(curp, specials);
				if (curp[n]) {
					if (n == 0) {
						buf[total++] = *curp++;
						continue;
					} else if (curp[n + 1] == 0) {
						if (n > max_size)
							n = max_size;
						memcpy(buf + total, curp, n);
						total += n;
						curp += n;
						continue;
					} else {
						buf[total++] = '"';
						state |= ST_QUOTE;
						continue;
					}
				} else
					state &= ~ST_QUOTE;
			}
			
			if ((state & ST_QUOTE) && strchr(escapable, *curp)) {
				if (total + 2 > max_size)
					break;
				buf[total++] = '\\';
				curp++;
			}
			buf[total++] = *curp++;
		}
		return total;
	} else if (is_interactive()) {
		int n;
		
		prompt();
		if (fgets(buf, max_size, yyin) == NULL)
			return 0;
		n = strlen(buf);
		need_prompt = (n > 0 && buf[n-1] == '\n');
		return n;
	}
	return fread(buf, 1, max_size, yyin);
}

void
lexer_from_file(char const *name)
{
	if (strcmp(name, "-") == 0)
		yyin = stdin;
	else {
		yyin = fopen(name, "r");
		if (!yyin)
			abend("can't open input file %s: %s",
			      name, strerror(errno));
	}
}

void
lexer_from_argv(int argc, char **argv)
{
        inargc = argc;
        inargv = argv;
        next = 0;
        curp = NULL;
}

#undef YY_INPUT
#define YY_INPUT(buf,result,max_size) result = fillbuf (buf, max_size)

%}
%x STR DOM
D [0-9]
WS [ \t]
C [^ \t\"\n;]
%%
nxdomain        return T_NXDOMAIN;
yxdomain        return T_YXDOMAIN;
nxrrset         return T_NXRRSET;
yxrrset         return T_YXRRSET;
del|delete      return T_DELETE;
add             return T_ADD;
update          return T_UPDATE;
send            return T_SEND;

\"[^\\"]*\"     { yylval.s = str_new(yytext + 1, yyleng - 2);
                  return T_STRING; }
\"[^\\"]*\\.    { BEGIN(STR);
                  str_start(yytext + 1, yyleng - 3);
                  str_esc(yytext[yyleng-1]); }
<STR>[^\\"]*\\. { str_add(yytext, yyleng - 2);
                  str_esc(yytext[yyleng-1]); }
<STR>[^\\"]*\"  { str_add(yytext, yyleng - 1);
                  yylval.s = str_stop();
                  BEGIN(INITIAL);
                  return T_STRING; }

<DOM>{C}+       { yylval.s = str_new(yytext, yyleng);
                  BEGIN(INITIAL);
                  return T_STRING; }
<DOM>.          { yyless(0); BEGIN(INITIAL); }

{D}+            { errno = 0;
                  yylval.n = strtoul(yytext, NULL, 10);
                  if (errno) {
                       abend("%s: %s", yytext, strerror(errno));
                  }
                  return T_NUMBER; }

{C}+            { yylval.s = str_new(yytext, yyleng);
                  return T_STRING; }

\n|;            return ';';

{WS}+ ;

\\.             { str_start(NULL, 0);
                  str_esc(yytext[1]);
                  yylval.s = str_stop();
                  return T_STRING; }
.               { abend("stray character %#o in input stream", yytext[0]); }
%%
int
yywrap(void)
{
    return 1;
}

void
need_domain(void)
{
    BEGIN(DOM);
}

enum {
	RDATA_PARSE_OK,
	RDATA_PARSE_NONE,
	RDATA_PARSE_ERR
};

static int
a_parser(nsu_rr *rr)
{
	struct addrinfo hints, *res;

	switch (yylex()) {
	case T_STRING:
		break;

	case ';':
		yyless(0);
		return RDATA_PARSE_NONE;

	case 0:
		return RDATA_PARSE_NONE;
		
	default:
		yyerror("unexpected token");
		yyless(0);
		return RDATA_PARSE_ERR;
	}	

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;
	
	if (getaddrinfo(yylval.s, NULL, &hints, &res)
	    || res->ai_addrlen != sizeof(struct sockaddr_in)) {
		abend("%s: invalid IP address", yylval.s);
		return RDATA_PARSE_ERR;
	}
	
	memcpy(&rr->data.d_a,
	       &((struct sockaddr_in*)res->ai_addr)->sin_addr,
	       sizeof(rr->data.d_a));
	freeaddrinfo(res);
	return RDATA_PARSE_OK;
}

static int
txt_parser(nsu_rr *rr)
{
	switch (yylex()) {
	case T_STRING:
		rr->data.d_txt = yylval.s;
		break;

	case ';':
		yyless(0);
		return RDATA_PARSE_NONE;

	case 0:
		return RDATA_PARSE_NONE;
		
	default:
		yyerror("unexpected token");
		yyless(0);
		return RDATA_PARSE_ERR;
	}
	return RDATA_PARSE_OK;
}

static int
domain_name_parser(nsu_rr *rr)
{
	need_domain();
	switch (yylex()) {
	case T_STRING:
		rr->data.d_name = yylval.s;
		break;

	case ';':
		yyless(0);
		return RDATA_PARSE_NONE;

	case 0:
		return RDATA_PARSE_NONE;
		
	default:
		yyerror("unexpected token");
		yyless(0);
		return RDATA_PARSE_ERR;
	}
	return RDATA_PARSE_OK;
}

static int
mx_parser(nsu_rr *rr)
{
	switch (yylex()) {
	case T_NUMBER:
		rr->data.d_mx.pref = yylval.n;
		break;

	case ';':
		yyless(0);
		return RDATA_PARSE_NONE;

	case 0:
		return RDATA_PARSE_NONE;
		
 	default:
		yyerror("unexpected token");
		yyless(0);
 		return RDATA_PARSE_ERR;
	}

	need_domain();
	
	switch (yylex()) {
	case T_STRING:
		rr->data.d_mx.name = yylval.s;
		break;

	case ';':
		yyless(0);
		yyerror("expected domain-name before ;");
		BEGIN(INITIAL);
		return RDATA_PARSE_ERR;

	case 0:
		yyerror("expected domain-name at the end of input");
		BEGIN(INITIAL);
		return RDATA_PARSE_ERR;
		     
 	default:
		abend("expected domain-name, but found %s", yytext);
		yyless(0);
		BEGIN(INITIAL);
 		return RDATA_PARSE_ERR;
	}

	return RDATA_PARSE_OK;
}
	
typedef int (*RDATA_PARSER)(nsu_rr *rr);

static RDATA_PARSER rdata_parser[] = {
	[ns_t_a] = a_parser,
	[ns_t_txt] = txt_parser,
	[ns_t_ns] = domain_name_parser,
	[ns_t_cname] = domain_name_parser,
	[ns_t_ptr] = domain_name_parser,
	[ns_t_mx] = mx_parser
};

int
rdata_parse(nsu_rr *rr, int type)
{
	memset(rr, 0, sizeof(*rr));
	rr->type = ns_t_any;
	if (type >=0 &&
	    type < sizeof(rdata_parser) / sizeof(rdata_parser[0]) &&
	    rdata_parser[type]) {
		int rc = rdata_parser[type](rr);
		if (rc == RDATA_PARSE_OK)
			rr->type = type;
		else if (rc == RDATA_PARSE_ERR)
			return -1;
		return 0;
	}
	yyerror("unsupported data type");
	return -1;
}
	    
		

