#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <arpa/nameser.h>
#include <resolv.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <nettle/md5.h>
#include <nettle/sha1.h>
#include <nettle/sha2.h>
#include <nettle/hmac.h>
#include <nettle/base64.h>
#include "nsu.h"
#include "nsu_app.h"

char const *progname;
unsigned long default_ttl = 600;
char const *domain_name;
struct sockaddr_in *server_addr;
char const domain_service[] = "53";
struct nsu_signer *tsig;

void
vdiagprt(char const *fmt, va_list ap)
{
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
}

void
panic(char const *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vdiagprt(fmt, ap);
	va_end(ap);
}

void
abend(char const *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	vdiagprt(fmt, ap);
	va_end(ap);
	exit(1);
}

void
abend_nomem(void)
{
	abend("not enough memory");
}

#define MAX_NS 20

static void
ns_perror(char const *name, int verbose)
{
	fprintf(stderr, "%s: %s: ", progname, name);
	fputc('\n', stderr);
}

static char *ns_rcode_strtab[] = {
	[ns_r_noerror] = "No error occurred.",
        [ns_r_formerr] = "Format error.",
        [ns_r_servfail] = "Server failure.",
        [ns_r_nxdomain] = "Name error.",
        [ns_r_notimpl] = "Unimplemented.",
        [ns_r_refused] = "Operation refused.",
        [ns_r_yxdomain] = "Name exists.",
        [ns_r_yxrrset] = "RRset exists.",
        [ns_r_nxrrset] = "RRset does not exist.",
        [ns_r_notauth] = "Not authoritative for zone.",
        [ns_r_notzone] = "Zone of record different from zone section.",
        [ns_r_max] = "ns_r_max",
        [ns_r_badvers] = "ns_r_badvers",
        [ns_r_badsig] = "Bad signature.",
        [ns_r_badkey] = "Bad key.",
        [ns_r_badtime] = "Bad time."
};

char const *
ns_rcodestr(int code)
{
	if (code >= 0 &&
	    code < sizeof(ns_rcode_strtab)/sizeof(ns_rcode_strtab[0]) &&
	    ns_rcode_strtab[code])
		return ns_rcode_strtab[code];
	return "Unknown response code";
}

void
nstab_add(ns_msg handle, ns_rr rr, struct sockaddr_in *nstab, int *nsnum)
{
	char name[MAXDNAME];
	struct addrinfo hints, *addr, *ap;
	
	if (ns_name_uncompress(ns_msg_base(handle),
			       ns_msg_end(handle),
			       ns_rr_rdata(rr),
			       name,
			       sizeof(name)) < 0) {
		abend("ns_name_uncompress failed");
	}

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;

	if (getaddrinfo(name, domain_service, &hints, &addr)) {
		abend("%s: invalid IP address", name);
	}
	
	for (ap = addr; ap && *nsnum < MAX_NS; ap = ap->ai_next) {
		if (ap->ai_addrlen == sizeof(struct sockaddr_in)) {
			int i;

			for (i = 0; i < *nsnum; i++) {
				if (!memcmp(&nstab[i], ap->ai_addr,
					    ap->ai_addrlen)) {
					goto next;
				}
			}
			memcpy(&nstab[*nsnum], ap->ai_addr, ap->ai_addrlen);
			++ *nsnum;
		}
	next:
		continue;
	}
	freeaddrinfo(addr);
}

void
ns_add(struct sockaddr_in *nstab, int *nsnum, ns_msg handle, ns_sect section)
{
	int rrnum;	  /* resource record number */

	for (rrnum = 0; rrnum < ns_msg_count(handle, section); rrnum++) {
		ns_rr rr;

		if (*nsnum == MAX_NS)
			break;

		if (ns_parserr(&handle, section, rrnum, &rr)) {
			panic("ns_parserr: %s", strerror(errno));
		}

		if (ns_rr_type(rr) == ns_t_ns) {
			nstab_add(handle, rr, nstab, nsnum);
		}
	}
}

typedef union {
	HEADER hdr;
	u_char buf[4096];
} QUERYBUF;


void
ns_add_master(char const *domain, struct sockaddr_in *nstab, int *nsnum)
{
	QUERYBUF response, query;
	int rlen, qlen;
	ns_msg handle;
	ns_rr rr;
	char name[MAXDNAME];
	
	qlen = res_mkquery(ns_o_query, /* regular query */
			   domain, /* the zone to look up */
			   ns_c_in, /* Internet type */
			   ns_t_soa, /* look up an SOA record */
			   NULL, /* always NULL */
			   0, /* length of NULL */
			   NULL, /* always NULL */
			   query.buf, /* buffer for the query */
			   sizeof(query));	/* size of the buffer */

	rlen = res_send(query.buf, /* the query */
			qlen, /* true length */
			response.buf, /*buffer */
			sizeof(response));

	if (rlen < 0) {
		if (errno == ECONNREFUSED) {	/* no server on the host */
			panic("no name servers configured");
		} else {
			/* anything else: no response */
			panic("no response");
		}
		return;
	}

	if (ns_initparse(response.buf, rlen, &handle) < 0) {
		panic("ns_initparse: %s", strerror(errno));
		return;
	}

	if (ns_msg_getflag(handle, ns_f_rcode) != ns_r_noerror) {
		panic("%s: %s", domain,
		       ns_rcodestr(ns_msg_getflag(handle, ns_f_rcode)));
		return;
	}

	if (ns_msg_count(handle, ns_s_an) != 1) {
		panic("%s: expected 1 answer, got %d",
		      domain, ns_msg_count(handle, ns_s_an));
		return;
	}

	if (ns_parserr(&handle, ns_s_an, 0, &rr)) {
		if (errno != ENODEV) {
			panic("ns_parserr: %s", strerror(errno));
		}
	}

	if (ns_rr_type(rr) != ns_t_soa) {
		panic("expected answer type %d, got %d",
		      ns_t_soa, ns_rr_type(rr));
		return;
	}

	nstab_add(handle, rr, nstab, nsnum);
}

static int
find_ns(char const *domain, struct sockaddr_in *nstab, int verbose)
{
	int nscount = 0;
	QUERYBUF response; /* response buffers */
	int resplen; /* buffer length */
	ns_msg handle;

	resplen = res_query(domain, ns_c_in, ns_t_ns,
			    response.buf, sizeof(response));
	if (resplen < 0) {
		switch (h_errno) {
		case HOST_NOT_FOUND:
			if (verbose)
				panic("%s: host not found", domain);
			return 0;

		case NO_DATA:
			if (verbose)
				panic("%s: no NS records", domain);
			return 0;

		case TRY_AGAIN:
			panic("no response for DNS query");
			break;

		default:
			panic("%s: unexpected error (%d)", h_errno);
		}
		return -1;
	}

	if (ns_initparse(response.buf, resplen, &handle) < 0) {
		panic("ns_initparse: %s", strerror(errno));
		return -1;
	}

	ns_add_master(domain, nstab, &nscount);
	ns_add(nstab, &nscount, handle, ns_s_an);
	ns_add(nstab, &nscount, handle, ns_s_ns);

	return nscount;
}

static void
set_server(char *name)
{
	struct addrinfo hints, *r;
	char *service;
	struct __res_state *st = &_res;

	service = strchr(name, ':');
	if (service)
		*service++ = 0;
	else
		service = (char*)domain_service;

	memset(&hints, 0, sizeof(hints));
	hints.ai_family = AF_INET;

	if (getaddrinfo(name, service, &hints, &r)
	    || r->ai_addrlen != sizeof(struct sockaddr_in)) {
		abend("%s: invalid IP address", name);
	}
	server_addr = malloc(r->ai_addrlen);
	if (!server_addr)
		abend_nomem();
	memcpy(server_addr, r->ai_addr, r->ai_addrlen);
	memcpy(&_res.nsaddr_list[0], r->ai_addr, r->ai_addrlen);
	_res.nscount = 1;
	freeaddrinfo(r);
}

static int
get_arg_ul(char const *arg, unsigned long *np)
{
	char *p;
	unsigned long n;
	errno = 0;
	n = strtoul(arg, &p, 10);
	if (errno || *p)
		return -1;
	*np = n;
	return 0;
}

struct tsig_alg {
	char *name;
	int code;
	size_t digest_size;
	size_t ctx_size;
	const struct nettle_hash *nhash;
};

char const *
tsig_alg_name(struct tsig_alg const *alg)
{
	return alg->name;
}

size_t
tsig_alg_digest_size(struct tsig_alg const *alg)
{
	return alg->digest_size;
}

enum { CTX_INNER, CTX_OUTER, CTX_STATE };

static inline void *
nsu_signer_ctx(struct nsu_signer *sgn, int n)
{
	return sgn->ctx + n * sgn->alg->ctx_size;
}

void
nsu_signer_set_key(struct nsu_signer *sgn, u_char const *key, size_t len)
{
	hmac_set_key(nsu_signer_ctx(sgn, CTX_OUTER),
		     nsu_signer_ctx(sgn, CTX_INNER),
		     nsu_signer_ctx(sgn, CTX_STATE),
		     sgn->alg->nhash,
		     len, key);
}
	
void
nsu_signer_update(struct nsu_signer *sgn, u_char const *data, size_t len)
{
	hmac_update(nsu_signer_ctx(sgn, CTX_STATE), sgn->alg->nhash, len, data);
}

void
nsu_signer_digest(struct nsu_signer *sgn, u_char *data, size_t len)
{
	hmac_digest(nsu_signer_ctx(sgn, CTX_OUTER),
		    nsu_signer_ctx(sgn, CTX_INNER),
		    nsu_signer_ctx(sgn, CTX_STATE),
		    sgn->alg->nhash,
		    len, data);
}

#define CODE_END 0
#define CODE_ALIAS (-1)

static struct tsig_alg algtab[] = {
	{ .name = "HMAC-MD5.SIG-ALG.REG.INT",
	  .code = 157,
	  .digest_size = MD5_DIGEST_SIZE,
	  .ctx_size = sizeof(struct md5_ctx),
	  .nhash = &nettle_md5,
	},
	{ .name = "HMAC-MD5",
	  .code = CODE_ALIAS,
	},
	{ .name = "HMAC-SHA1",
	  .code = 161,
	  .digest_size = SHA1_DIGEST_SIZE,
	  .nhash = &nettle_sha1,
	  .ctx_size = sizeof(struct sha1_ctx),
	},
	{ .name = "HMAC-SHA",
	  .code = CODE_ALIAS
	},
	{ .name = "HMAC-SHA224",
	  .code = 162,
	  .digest_size = SHA224_DIGEST_SIZE,
	  .nhash = &nettle_sha224,
	  .ctx_size = sizeof(struct sha224_ctx),
	},
	{ .name = "HMAC-SHA256",
	  .code = 163,
	  .digest_size = SHA256_DIGEST_SIZE,
	  .nhash = &nettle_sha256,
	  .ctx_size = sizeof(struct sha256_ctx),
	},
	{ .name = "HMAC-SHA384",
	  .code = 164,
	  .digest_size = SHA384_DIGEST_SIZE,
	  .nhash = &nettle_sha384,
	  .ctx_size = sizeof(struct sha384_ctx),
	},
	{ .name = "HMAC-SHA512",
	  .code = 165,
	  .digest_size = SHA512_DIGEST_SIZE,
	  .nhash = &nettle_sha512,
	  .ctx_size = sizeof(struct sha512_ctx),
	},
	{ .code = CODE_END }
};

static struct tsig_alg const *
alg_find_code(int code)
{
	struct tsig_alg *ap;
	for (ap = algtab; ap->code != CODE_END; ap++)
		if (ap->code == code)
			return ap;
	return NULL;
}

static struct tsig_alg const *
alg_find_name(char const *name)
{
	struct tsig_alg *ap;

	for (ap = algtab; ap->code == CODE_END; ap++) {
		if (strcasecmp(ap->name, name) == 0)
			break;
	}

	while (ap->code == CODE_ALIAS && ap > algtab)
		ap--;

	if (ap->code <= 0)
		return NULL;

	return ap;
}

static int
is_suffix(char const *suf, char const *str)
{
	size_t suflen = strlen(suf);
	size_t len = strlen(str);
	return suflen < len && memcmp(str + len - suflen, suf, suflen) == 0;
}

static char const *priv_suffix = ".private";
static char const *key_suffix = ".key";

static struct tsig_alg const *
keyfile_read_private(char const *name, uint8_t **pkey, size_t *pkey_len)
{
	char *tmpname = NULL;
	FILE *fp;
	char buf[512];
	int lno = 0;

	struct tsig_alg const *alg = NULL;
	uint8_t *key;
	size_t key_len;
	
	static char algorithm_field[] = "Algorithm";
	static int algorithm_length = sizeof(algorithm_field) - 1;
	static char key_field[] = "Key";
	static int key_length = sizeof(key_field) - 1;

	if (is_suffix(key_suffix, name)) {
		size_t baselen = strlen(name) - strlen(key_suffix);
		tmpname = malloc(baselen + strlen(priv_suffix) + 1);
		if (!tmpname)
			abend_nomem();
		memcpy(tmpname, name, baselen);
		strcpy(tmpname + baselen, priv_suffix);
		name = tmpname;
	} else if (access(name, F_OK) && errno == ENOENT) {
		size_t namelen = strlen(name);
		if (name[namelen-1] == '.') {
			tmpname = malloc(namelen + strlen(priv_suffix));
			if (!tmpname)
				abend_nomem();
			memcpy(tmpname, name, namelen - 1);
			strcpy(tmpname + namelen - 1, priv_suffix);
			name = tmpname;
		}
	}

	fp = fopen(name, "r");
	if (!fp)
		abend("can't open %s: %s", name, strerror(errno));
	while (fgets(buf, sizeof(buf), fp)) {
		size_t len, n;
		lno++;
		len = strlen(buf);
		if (len == 0)
			break;
		if (buf[len-1] != '\n')
			abend("%s:%d: line too long", name, lno);
		buf[--len] = '\n';
		n = strcspn(buf, ":");
		if (buf[n] == 0)
			abend("%s:%d: malformed line", name, lno);
		if (n == algorithm_length &&
		    memcmp(algorithm_field, buf, n) == 0) {
			char *p;
			long code;
			errno = 0;
			code = strtol(buf + n + 1, &p, 10);
			if (errno || *p != ' ')
				abend("%s:%d: malformed line", name, lno);
			
			alg = alg_find_code(code);
			if (!alg)
				abend("%s:%d: unsupported algorithm %s",
				      name, lno, buf + n);
		} else if (n == key_length && memcmp(key_field, buf, n) == 0) {
			struct base64_decode_ctx ctx;
			size_t src_len;
			char *p;

			p = buf + n + 1;
			while (*p == ' ' || *p == '\t')
				p++;
			src_len = strlen(p);
			key_len = BASE64_DECODE_LENGTH(src_len);
			key = malloc(key_len);
			if (!key)
				abend_nomem();
			
			base64_decode_init(&ctx);
			base64_decode_update(&ctx, &key_len, key, src_len, p);
			if (!base64_decode_final(&ctx))
				abend("%s:%d: bad base64", name, lno);
		}
	}
	fclose(fp);

	if (!alg)
		abend("%s:%d: no %s clause", algorithm_field);
	if (!key)
		abend("%s:%d: no %s clause", key_field);

	free(tmpname);
	
	*pkey = key;
	*pkey_len = key_len;
	return alg;
		
}

static char *
keyfile_read_key(char const *name, uint8_t **pkey, size_t *pkey_len)
{
	char *tmpname = NULL;
	char buf[512];
	char *ret, *keyptr;
	size_t n;
	uint8_t *key;
	FILE *fp;
	int i;
	size_t src_len, key_len;
	struct base64_decode_ctx ctx;
	
	if (is_suffix(priv_suffix, name)) {
		size_t baselen = strlen(name) - strlen(priv_suffix);
		tmpname = malloc(baselen + strlen(key_suffix) + 1);
		if (!tmpname)
			abend_nomem();
		memcpy(tmpname, name, baselen);
		strcpy(tmpname + baselen, key_suffix);
		name = tmpname;
	} else if (access(name, F_OK) && errno == ENOENT) {
		size_t namelen = strlen(name);
		if (name[namelen-1] == '.') {
			tmpname = malloc(namelen + strlen(key_suffix));
			if (!tmpname)
				abend_nomem();
			memcpy(tmpname, name, namelen - 1);
			strcpy(tmpname + namelen - 1, key_suffix);
			name = tmpname;
		}
	}
	
	fp = fopen(name, "r");
	if (!fp)
		abend("can't open %s: %s", name, strerror(errno));
	if (!fgets(buf, sizeof(buf), fp))
		abend("%s: read error", name);
	fclose(fp);
	
	n = strcspn(buf, " ");
	if (strncmp(buf + n, " IN KEY", 7))
		abend("%s:%d: malformed line", name, 1);

	keyptr = buf + n + 7;
	for (i = 0; i < 3; i++) {
		while (*keyptr && isspace(*keyptr))
			keyptr++;
		while (*keyptr && isdigit(*keyptr))
			keyptr++;
	}
	while (*keyptr && isspace(*keyptr))
		keyptr++;

	src_len = strlen(keyptr);
	key_len = BASE64_DECODE_LENGTH(src_len);
	key = malloc(key_len);
	if (!key)
		abend_nomem();
			
	base64_decode_init(&ctx);
	base64_decode_update(&ctx, &key_len, key, src_len, keyptr);
	if (!base64_decode_final(&ctx))
		abend("%s:%d: bad base64", name, 1);
	*pkey = key;
	*pkey_len = key_len;
	
	if (buf[n-1] == '.')
		--n;
	ret = malloc(n+1);
	if (!ret)
		abend_nomem();
	memcpy(ret, buf, n);
	ret[n] = 0;

	free(tmpname);
	
	return ret;
}

static struct nsu_signer *
keyfile_read(char const *name)
{
	struct tsig_alg const *alg;
	uint8_t *pkey, *key;
	size_t pkey_len, key_len;
	char *keyname;
	struct nsu_signer *sgn;
	
	alg = keyfile_read_private(name, &pkey, &pkey_len);
	keyname = keyfile_read_key(name, &key, &key_len);

	if (pkey_len != key_len || memcmp(pkey, key, key_len))
		abend("%s: keys don't match", name);
	free(pkey);
	
	sgn = calloc(1, sizeof(*sgn));
	if (!sgn)
		abend_nomem();
	sgn->ctx = malloc(3 * alg->ctx_size);
	if (!sgn->ctx)
		abend_nomem();
	sgn->digest = malloc(tsig_alg_digest_size(alg));
	if (!sgn->digest)
		abend_nomem();
	
	sgn->name = keyname;
	sgn->alg = alg;
	sgn->fudge = 300; //FIXME

	nsu_signer_set_key(sgn, key, key_len);
	free(key);

	return sgn;
}

void
nsu_signer_free(struct nsu_signer *sgn)
{
	if (!sgn)
		return;
	free(sgn->name);
	free(sgn->ctx);
	free(sgn->digest);
	free(sgn);
}

/*
 * nsu
 */
int
main(int argc, char **argv)
{
	char *input_file = NULL;
	char *keyfile_name = NULL;
	int i;
	nsu_rr upd;
	char *p;

	progname = argv[0];
	res_init();
	yy_flex_debug = 0;
	while ((i = getopt(argc, argv, "d:f:k:s:t:x")) != EOF) {
		switch (i) {
		case 'd':
			domain_name = optarg;
			break;
			
		case 'f':
			input_file = optarg;
			break;

		case 'k':
			keyfile_name = optarg;
			break;
			
		case 's':
			set_server(optarg);
			break;

		case 't':
			if (get_arg_ul(optarg, &default_ttl))
				exit(1);
			break;

		case 'x':
			if (yydebug == 1)
				yy_flex_debug = 1;
			else
				yydebug = 1;
			break;
			
		default:
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;

	if (keyfile_name)
		tsig = keyfile_read(keyfile_name);
	
	if (input_file)
		lexer_from_file(input_file);
	else if (argc > 0)
		lexer_from_argv(argc, argv);
	else
		lexer_from_file("-");
	yyparse();
	exit(0);
}

enum {
	update_fail = -1,
	update_ok,
	update_retry
};

int
send_update_query(res_state state, char const *zone, u_char *qbuf, int qlen)
{
	union {
		HEADER hdr;
		u_char buf[NS_PACKETSZ*2];
	} response;
	int rlen;
	ns_msg handle;
	int rcode;
	
	if (state->options & RES_DEBUG) {
		char name[NS_MAXDNAME];
		
		getnameinfo((struct sockaddr*)state->nsaddr_list,
			    sizeof(state->nsaddr_list[0]),
			    name, sizeof(name),
			    NULL, 0,
			    NI_NUMERICHOST);
		printf(";; trying nameserver %s\n", name);
	}

	rlen = res_nsend(state, qbuf, qlen,
			 (u_char *) &response,
			 sizeof(response));
	
	if (rlen < 0) {
		if (errno == ECONNREFUSED) {	/* no server on the host */
			panic("connection refused");
		} else {
			/* anything else: no response */
			panic("no response");
		}
		return update_fail;
	}

	if (ns_initparse(response.buf, rlen, &handle) < 0) {
		panic("ns_initparse: %s", strerror(errno));
		return update_fail;
	}

	rcode = ns_msg_getflag(handle, ns_f_rcode);
	switch (rcode) {
/*
 * If a response is received whose RCODE is SERVFAIL or NOTIMP, or
 * if no response is received within an implementation dependent timeout
 * period, or if an ICMP error is received indicating that the server's
 * port is unreachable, then the requestor will delete the unusable
 * server from its internal name server list and try the next one,
 * repeating until the name server list is empty.  If the requestor runs
 * out of servers to try, an appropriate error will be returned to the
 * requestor's caller.
 */
	case ns_r_notimpl:
	case ns_r_servfail:
		if (state->options & RES_DEBUG) {
			char name[NS_MAXDNAME];

			getnameinfo((struct sockaddr*)state->nsaddr_list,
				    sizeof(state->nsaddr_list[0]),
				    name, sizeof(name),
				    NULL, 0,
				    NI_NUMERICHOST);
			printf(";; nameserver %s: %s", name,
			       ns_rcodestr(rcode));
		}
		return update_retry;
		
	case ns_r_noerror:
		break;
		
	default:
		panic("%s: %s", zone, ns_rcodestr(rcode));
		return update_fail;
	}
	return update_ok;
}

int
run_update(nsu_rr *prereq, int num_prereq,
	   nsu_rr *update, int num_update)
{
	struct sockaddr_in nstab[MAX_NS];
	int nscount = 0;
	int qlen;
	QUERYBUF query;
	char const *zone = NULL;
	int rc;
	int i;
	struct __res_state rstate;
	
	if (!domain_name) {
		char *p = update->name;

		while ((p = strchr(p, '.')) != NULL) {
			p++;
			nscount = find_ns(p, nstab, 0);
			if (nscount > 0) {
				zone = p;
				break;
			}
		}
		if (!zone)
			abend("cannot determine nameservers for update");
	} else {
		zone = domain_name;
		nscount = find_ns(zone, nstab, 1);
		if (nscount == 0)
			return -1;
	}

#if 0
	/* Override the name servers */
	if (server_addr) {
		memcpy(nstab, server_addr, sizeof(nstab[0]));
		nscount = 1;
	}
#endif
	qlen = nsu_mkquery(zone,
			   prereq, num_prereq,
			   update, num_update,
			   tsig,
			   query.buf, sizeof(query.buf));
	
	if (qlen < 0)
		abend("failed to create update query");


	res_ninit(&rstate);
	rstate.nscount = 1;
	for (i = 0; i < nscount; i++) {
		memcpy(&rstate.nsaddr_list[0], nstab + i,
		       sizeof(rstate.nsaddr_list[0]));
		rc = send_update_query(&rstate, zone, query.buf, qlen);
		if (rc != update_retry)
			break;
	}
	return rc == update_ok ? 0 : -1;
}
